from functools import partial

import numpy as np
from scipy.sparse import issparse, coo_matrix

import io

# flake8: noqa: E402
from nqxpack._src.lib_v1.custom_types import (
    register_serialization,
    register_automatic_serialization,
)
from nqxpack._src.contextmgr import current_context


def _pack_array(array):
    if isinstance(array, np.ndarray):
        return array
    elif issparse(array):
        array = array.tocoo()
        return {
            "_target_": "scipy.sparse.coo_matrix",
            "data": array.data,
            "row": array.row,
            "col": array.col,
            "shape": list(array.shape),
        }
    else:
        raise ValueError(f"Unsupported array type: {type(array)}")


def _unpack_array(array, dtype=None):
    if isinstance(array, np.ndarray):
        if dtype is not None:
            array = array.astype(dtype)
        return array
    elif isinstance(array, dict) and array.get("_target_") == "scipy.sparse.coo_matrix":
        return coo_matrix(
            (array["data"], (array["row"], array["col"])),
            shape=array["shape"],
            dtype=dtype,
        )


## LocalOperator
from netket.operator import LocalOperator, LocalOperatorJax


def serialize_LocalOperator(op):

    operators = [_pack_array(term) for term in op.operators]
    current_context().asset_manager.write_msgpack("operators.msgpack", operators)

    if hasattr(op, "_acting_on"):
        acting_on = op._acting_on.tolist()
    else:
        acting_on = [list(int(o) for o in ao) for ao in op.acting_on]

    return {
        "hilbert": op.hilbert,
        # "operators": op.operators,
        "acting_on": acting_on,
        "constants": op.constant.item(),
        "dtype": op.dtype,
    }


def deserialize_LocalOperator(cls, obj):

    operators = current_context().asset_manager.read_msgpack("operators.msgpack")
    operators = [_unpack_array(arr, dtype=obj["dtype"]) for arr in operators]

    return cls(
        hilbert=obj["hilbert"],
        operators=operators,
        acting_on=obj["acting_on"],
        constant=obj["constants"],
        dtype=obj["dtype"],
    )


register_serialization(
    LocalOperator,
    serialize_LocalOperator,
    partial(deserialize_LocalOperator, LocalOperator),
)
register_serialization(
    LocalOperatorJax,
    serialize_LocalOperator,
    partial(deserialize_LocalOperator, LocalOperatorJax),
)

## PauliStrings
from netket.operator import PauliStrings, PauliStringsJax


def serialize_PauliStrings(op):
    ctx = current_context()
    asset_manager = ctx.asset_manager

    buffer = io.BytesIO()
    np.savez_compressed(buffer, operators=op.operators, weights=op.weights)
    asset_manager.write_asset("data.npz", buffer.getvalue())

    return {
        "hilbert": op.hilbert,
        # "operators": op.operators,
        # "weights": op._weights,
        "cutoff": op._cutoff,
        "dtype": op.dtype,
    }


def deserialize_PauliStrings(cls, obj):
    ctx = current_context()
    asset_manager = ctx.asset_manager

    data = np.load(io.BytesIO(asset_manager.read_asset("data.npz")))
    operators = data["operators"]
    weights = data["weights"]

    return cls(
        hilbert=obj["hilbert"],
        operators=operators,
        weights=weights,
        cutoff=obj["cutoff"],
        dtype=obj["dtype"],
    )


register_serialization(
    PauliStrings,
    serialize_PauliStrings,
    partial(deserialize_PauliStrings, PauliStrings),
)
register_serialization(
    PauliStringsJax,
    serialize_PauliStrings,
    partial(deserialize_PauliStrings, PauliStringsJax),
)

## Ising
from netket.operator import Ising, IsingJax


def serialize_Ising(op):
    return {
        "hilbert": op.hilbert,
        "graph": op._edges,
        "h": op.h,
        "J": op.J,
        "dtype": op.dtype,
    }


register_automatic_serialization(Ising, "hilbert", "h", "J", "dtype", graph="edges")
register_automatic_serialization(IsingJax, "hilbert", "h", "J", "dtype", graph="edges")



from functools import partial

from nqxpack._src.lib_v1.custom_types import register_serialization


def serialize_partial(par):
    res = {
        "name": par.func,
    }
    if par.args:
        res["args"] = par.args
    if par.keywords:
        res["kwargs"] = par.keywords
    return res


def deserialize_partial(obj):
    return partial(obj["name"], *obj.get("args", ()), **obj.get("kwargs", {}))


register_serialization(partial, serialize_partial, deserialize_partial)


# frozenset
def serialize_frozenset(obj):
    return {"elements": list(obj)}


def deserialize_frozenset(obj):
    return frozenset(obj["elements"])


register_serialization(frozenset, serialize_frozenset, deserialize_frozenset)



__all__ = [
    "serialize_object",
    "deserialize_object",
    "registry",
]
from nqxpack._src.lib_v1 import serialize_object, deserialize_object
from nqxpack._src.api import save, load

from nqxpack import registry



from pathlib import Path

from zipfile import ZipFile
import shutil


from nqxpack._src.lib_v1.asset_lib import ArchiveAssetManager, FolderAssetManager
from nqxpack._src.distributed import is_master_process


class ZipArchive:
    """
    Asset manager that writes to a zip archive.
    """

    def __init__(
        self,
        path,
        *args,
        mode="w",
        **kwargs,
    ):
        """
        Constructs an asset manager backed by a zip file archive.

        Args:
            archive: an open zip file object.
            remove_root: A prefix to remove from all keys when writing to the archive. This is optional.
            path: A prefix to add to all keys when writing to the archive. This is optional.
        """
        if not isinstance(path, Path):
            path = Path(path)

        if mode == "w" and not is_master_process():
            self._archive = None
        else:
            # Remove the directory and all its contents
            if path.exists() and path.is_dir():
                shutil.rmtree(path)
            self._archive = ZipFile(path, *args, mode=mode, **kwargs)

    def __enter__(self):
        if self._archive is None:
            return self
        self._archive.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._archive is None:
            return
        return self._archive.__exit__(exc_type, exc_value, traceback)

    def __contains__(self, key):
        return key in self._archive.namelist()

    def open(self, key, mode):
        return self._archive.open(key, mode)

    def read(self, key):
        return self._archive.read(key)

    def write(self, key, value):
        return self._archive.write(key, value)

    def close(self):
        self._archive.close()

    def create_asset_manager(self, *args, **kwargs):
        return ArchiveAssetManager(self._archive, *args, **kwargs)


class DirectoryArchive:
    """
    Asset manager that writes to a zip archive.
    """

    def __init__(self, path, mode="w"):
        """
        Constructs an asset manager backed by a zip file archive.

        Args:
            archive: an open zip file object.
            remove_root: A prefix to remove from all keys when writing to the archive. This is optional.
            path: A prefix to add to all keys when writing to the archive. This is optional.
        """
        self.path = Path(path)
        if is_master_process():
            if mode == "w":
                # Remove the directory and all its contents
                if self.path.exists() and self.path.is_dir():
                    shutil.rmtree(self.path)
                elif self.path.exists():
                    self.path.unlink()
                self.path.mkdir(parents=True, exist_ok=True)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return

    def __contains__(self, key):
        return (self.path / key).exists()

    def open(self, key, mode):
        mode = mode + "b"
        path = self.path / key
        if not path.parent.exists():
            path.parent.mkdir(parents=True)

        return path.open(mode)

    def read(self, key):
        path = self.path / key

        return path.read_bytes()

    def write(self, key, value):
        path = self.path / key
        if not path.parent.exists():
            path.parent.mkdir(parents=True)

        return path.write(value)

    def close(self):
        self.path.close()

    def create_asset_manager(self, *args, **kwargs):
        return FolderAssetManager(self.path, *args, **kwargs)



import numpy as np
from pathlib import Path

import jax
import jax.numpy as jnp
from flax import nnx
from flax import linen

import pytest

import nqxpack
import netket as nk
from netket_pro import distributed

hi = nk.hilbert.Spin(0.5, 4)
g = nk.graph.Chain(4)

models = {}
models["RBM"] = nk.models.RBM(param_dtype=complex)
models["SymmSymExp"] = nk.nn.blocks.SymmExpSum(
    nk.models.RBM(param_dtype=complex), symm_group=g.translation_group()
)


def build_nnx_model():
    return nnx.Sequential(
        nnx.Linear(
            in_features=4,
            out_features=2,
            rngs=nnx.Rngs(1),
        ),
        # nnx.relu,
        nnx.Linear(
            in_features=2,
            out_features=1,
            rngs=nnx.Rngs(1),
        ),
        jnp.squeeze,
    )


models["NNX-Sequential"] = build_nnx_model

models_params = [pytest.param(model, id=name) for name, model in models.items()]


@pytest.mark.parametrize("model", models_params)
def test_save_mcstate(model, tmpdir):
    if distributed.mode() == "sharding":
        tmpdir = Path(distributed.broadcast_string(str(tmpdir)))
    elif distributed.mode() == "mpi":
        tmpdir = nk.utils.mpi.MPI_py_comm.bcast(str(tmpdir), root=0)
        tmpdir = Path(tmpdir)

    if not isinstance(model, linen.Module):
        model = model()
    sa = nk.sampler.MetropolisLocal(hi, n_chains=4)
    vs = nk.vqs.MCState(sa, model, n_samples=64)

    distributed.barrier("barrier 1")
    nqxpack.save(vs, tmpdir / "mcstate.mpack")
    nqxpack.save(vs, "mcstate.mpack")
    distributed.barrier("barrier 2")

    new_vs = nqxpack.load(tmpdir / "mcstate.mpack")

    assert vs.hilbert == new_vs.hilbert
    jax.tree.map(np.testing.assert_allclose, vs.parameters, new_vs.parameters)
    jax.tree.map(
        np.testing.assert_allclose,
        distributed.allgather(vs.samples),
        distributed.allgather(new_vs.samples),
    )



from nqxpack._src.lib_v1.closure import register_closure_simple_serialization
from nqxpack._src.lib_v1.custom_types import register_serialization

import jax

register_closure_simple_serialization(
    jax.nn.initializers.normal,
    "init",
    original_qualname="jax._src.nn.initializers.normal",
)
register_closure_simple_serialization(
    jax.nn.initializers.variance_scaling,
    "init",
    original_qualname="jax._src.nn.initializers.variance_scaling",
)


def serialize_PyTreeDef(obj):
    return {
        "node_data": obj.node_data(),
        "children": obj.children(),
    }


def deserialize_PyTreeDef(obj):
    return jax.tree_util.PyTreeDef.make_from_node_data_and_children(
        jax.tree_util.default_registry, obj["node_data"], obj["children"]
    )


register_serialization(
    jax.tree_util.PyTreeDef, serialize_PyTreeDef, deserialize_PyTreeDef
)



import numpy as np

import jax.numpy as jnp

import netket as nk


def central_diff_grad(func, x, eps, *args, dtype=None):
    if dtype is None:
        dtype = x.dtype

    grad = np.zeros(
        len(x), dtype=nk.jax.maybe_promote_to_complex(x.dtype, func(x, *args).dtype)
    )
    epsd = np.zeros(len(x), dtype=dtype)
    epsd[0] = eps
    for i in range(len(x)):
        assert not np.any(np.isnan(x + epsd))
        grad_r = 0.5 * (func(x + epsd, *args) - func(x - epsd, *args))
        if jnp.iscomplexobj(x):
            grad_i = 0.5 * (func(x + 1j * epsd, *args) - func(x - 1j * epsd, *args))
            grad[i] = grad_r + 1j * grad_i
        else:
            grad[i] = grad_r

        assert not np.isnan(grad[i])
        grad[i] /= eps
        epsd = np.roll(epsd, 1)
    return grad


def same_derivatives(der, num_der, abs_eps=1.0e-6, rel_eps=1.0e-6):
    """
    Checks that two complex-valued arrays are the same.
    Same as `np.testing.assert_allclose` but checks the real
    and imaginary parts independently for better error reporting.
    """
    assert der.shape == num_der.shape

    np.testing.assert_allclose(der.real, num_der.real, rtol=rel_eps, atol=abs_eps)

    np.testing.assert_allclose(der.imag, num_der.imag, rtol=rel_eps, atol=abs_eps)



class MainScopeError(Exception):
    def __init__(self, name):
        super().__init__(
            f"""
            Impossible to serialize an object of type {name}.

            You cannot serialize objects defined in the main scope.
            Please define them in a module, and import the module.

            Of course, you will have to keep the module accessible to reload this file
            """
        )


class SerializationError(Exception):
    def __init__(self, typename, obj, path):
        super().__init__(
            f"""
            Impossible to serialize an object of type `{typename}` found at path
            `{path}`.

            You need to implement a custom serialization function for this object which
            converts in built in types or other serializable types, such as lists, integers
            dictionaries...

            To do it, you should add a new entry to the registry in the file
            in nqxpack._src.lib_v1.registry.

                from nqxpack._src.lib_v1 import (
                    register_serialization,
                    )

                def serialization_function(obj: typename) -> serializable_type:
                    ...

                register_serialization(
                    {typename}, serialization_function,
                    )

            """
        )


class FutureVersionError(Exception):
    def __init__(self, file_version, max_version):
        super().__init__(
            f"""
            File was saved with a more recent version of nqxpack/netket_pro than your current installation:

                File version      : {file_version}
                Supported version : {max_version}

            Please update the library to a version that supports this file format.
            """
        )



from nqxpack._src.lib_v1.custom_types import (
    register_automatic_serialization,
)

# Graph

# flax.nnx
from flax.nnx.graph import HashableMapping

register_automatic_serialization(HashableMapping, mapping="_mapping")



import orjson
import json

from pathlib import Path


from nqxpack._src.lib_v1 import (
    serialize_object,
    deserialize_object,
)
from nqxpack._src import distributed
from nqxpack._src.contextmgr import PackingContext
from nqxpack._src.metadata.generate_metadata import generate_metadata

from nqxpack._src.io import DirectoryArchive, ZipArchive
from nqxpack._src.errors import FutureVersionError
from nqxpack._src.lib_v1 import VERSION as LIB_VERSION
from nqxpack._src.registry import VERSION as REGISTRY_VERSION
from nqxpack._src.metadata.generate_metadata import VERSION as METADATA_VERSION

_CONFIG_FILENAME = "object.json"
_METADATA_FILENAME = "metadata.json"

_FORMAT_VERSION = 1.3

versioninfo_info = {
    "lib": LIB_VERSION,
    "registry": REGISTRY_VERSION,
    "metadata": METADATA_VERSION,
}


def save(object, path, *, zip: bool = True):
    """
    Saves an object to a file, using the NQXPack format.

    .. warning::

        The file will record the versions of installed packages whose objects are serialized to the file.
        A compatible version of those libraries will be required to load the file. By 'compatible', it is
        intended a version of libraries that is API-compatible with the version that was used to save the file.

        In general, we recomend to use the same version of the libraries that were used to save the file.

    The NQXPack format is a zip file that contains the following files:
    - metadata.json: Contains metadata about the file, including the format version and the versions of the
        libraries used to save the file.
    - object.json: A json file, in a format similar to Hydra's, that contains the information necessary to
        reconstruct the object.
    - assets/: A directory containing any object that cannot easily be serialized to json. This will generally
        contain large objects, like numpy arrays, stored into a binary blob.

    .. warning::

        The NQXPack format is in beta not guaranteed to be stable at the moment.

        It attempts to be backwards-compatible, but not forwards-compatible. This means that you can load
        files saved with older versions of NQXPack, but you cannot load files saved with newer versions. An
        informative error will be thrown in that case.
        However, the guarantee is, for the time-being, best-effort only.

        The NQXPack format is not secure. Do not load files from untrusted sources!

    Args:
        object: The object to save.
        path: The path to save the object to. If the path does not have a .nk extension, it will be added.
        zip: If True (default), the object will be saved in a zip file. If False, it will be saved in a directory.
    """
    if not isinstance(path, Path):
        path = Path(path)

    # Check that extension is .nk
    if path.suffix != ".nk":
        path = path.with_suffix(".nk")

    if zip:
        archive = ZipArchive(path, mode="w")
    else:
        archive = DirectoryArchive(path, mode="w")

    orjson_options = orjson.OPT_APPEND_NEWLINE | orjson.OPT_INDENT_2
    with archive:
        with PackingContext(
            asset_manager=archive.create_asset_manager(path="assets/"),
        ):
            object_json = serialize_object(object)

            if distributed.is_master_process():
                with archive.open(_CONFIG_FILENAME, "w") as f:
                    data = orjson.dumps(object_json, option=orjson_options)
                    f.write(data)

            metadata_json = generate_metadata(
                format_version=_FORMAT_VERSION, versioninfo=versioninfo_info
            )
            if distributed.is_master_process():
                with archive.open(_METADATA_FILENAME, "w") as f:
                    f.write(orjson.dumps(metadata_json, option=orjson_options))


def load(path):
    """
    Loads an nqxpack file.

    Args:
        object: The object to save.
        path: The path to save the object to. If the path does not have a .nk extension, it will be added.
        zip: If True (default), the object will be saved in a zip file. If False, it will be saved in a directory.
    """

    if not isinstance(path, Path):
        path = Path(path)

    if not path.exists() and path.suffix != ".nk":
        new_path = path.with_suffix(".nk")
        if new_path.exists():
            path = new_path

    if path.is_dir() and (path / _METADATA_FILENAME).exists():
        archive = DirectoryArchive(path, mode="r")
    elif path.exists():
        archive = ZipArchive(path, mode="r")
    else:
        raise FileNotFoundError(f"File not found: {path}")

    with archive:
        if _METADATA_FILENAME not in archive:
            raise ValueError(
                f"Invalid file format: archive does not contain {_METADATA_FILENAME}"
            )

        metadata = archive.read(_METADATA_FILENAME)
        metadata = json.loads(metadata)

        # validate
        if metadata["format"] != "NetKet":
            raise ValueError("Invalid file format.")

        file_version = metadata["format_version"]
        if file_version > _FORMAT_VERSION:
            # Do not attempt to load files from future versions
            raise FutureVersionError(file_version, _FORMAT_VERSION)
        elif file_version <= 1.1:
            # v1.0 and v1.1 of the format did not have the asset manager, and where just storing the state object
            # directly in the zip file.
            # Also, the object was stored in a file named config.json, not object.json
            _CONFIG_FILENAME = "config.json"
            asset_manager = archive.create_asset_manager(remove_root="state/")
        else:
            _CONFIG_FILENAME = "object.json"
            asset_manager = archive.create_asset_manager(path="assets/")

        config = archive.read(_CONFIG_FILENAME)
        with PackingContext(asset_manager=asset_manager, metadata=metadata):

            state_obj_dict = json.loads(config)

            # Fix a but that appeared briefly
            # TODO: Put this somewhere reasoanbly
            if "_target_" in state_obj_dict:
                if state_obj_dict["_target_"] == "netket.vqs.mc.mc_state.state.MCState":
                    state_obj_dict["_target_"] = "#netket.vqs.mc.mc_state.state.MCState"

            state = deserialize_object(state_obj_dict)

    return state



import re

from nqxpack._src.errors import (
    MainScopeError,
)
from nqxpack._src.contextmgr import current_context


def _register_package_version(module_name):
    import importlib.metadata

    # Extract the root package name (everything before the first dot)
    package_name = module_name.split(".")[0]

    PACKAGE_VERSIONS = current_context().package_versions

    if package_name in PACKAGE_VERSIONS:
        return PACKAGE_VERSIONS[package_name]

    # Look up the version of the root package
    try:
        version = importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        print(
            f"Package '{package_name}' is not installed, so you may not be able to reload this."
        )
        version = ""

    PACKAGE_VERSIONS[package_name] = version
    return version


def _qualname(obj, skip_register: bool = False):
    cls = obj if isinstance(obj, type) else type(obj)
    qname = cls.__module__ + "." + cls.__qualname__
    if cls.__module__.startswith("__main__"):
        raise MainScopeError(qname)

    if not skip_register:
        _register_package_version(cls.__module__)

    return qname


def _fname(fun, skip_register: bool = False):
    qname = fun.__module__ + "." + fun.__qualname__
    if fun.__module__.startswith("__main__"):
        raise MainScopeError(qname)

    if not skip_register:
        _register_package_version(fun.__module__)

    return qname


def _resolve_qualname(name: str):
    # this is to support parametric types
    match = re.match(r"([^\[]+)(?:\[(.+)\])?", name)
    if not match:
        raise ValueError(f"Invalid format: {name}")

    qualified_name = match.group(1)  # The part before the parametric brackets
    parametric_part = match.group(2)  # The content inside the brackets, if it exists

    if parametric_part is None:
        return _resolve_qualname_noparametric(qualified_name)
    else:
        qualified_name = _resolve_qualname_noparametric(qualified_name)
        parametric_part = _resolve_qualname_noparametric(parametric_part)
        return qualified_name[qualified_name]


def _resolve_qualname_noparametric(name: str):
    """
    Resolves a qualified name, guaranted not to have a parametric component.
    """

    # Split the qualified name into module and class
    parts = name.split(".")
    module_name = ".".join(parts[:-1])
    cls = parts[-1]

    module = None

    # Try to import it directly
    try:
        module = __import__(module_name, fromlist=[cls])
    except ImportError:
        pass
        # raise ImportError(
        #    f"Cannot import {module_name} necesary to resolve {name}. You need to install it to load this object."
        # )

    # it failed, maybe this is a path of the form `module.class.method`
    if module is None and len(parts) > 2:
        module_name = ".".join(parts[:-2])
        _cls = parts[-2]
        try:
            module = __import__(module_name, fromlist=[cls])
        except ImportError:
            raise ImportError(
                f"Cannot import {module_name} necesary to resolve {name}. You need to install it to load this object."
            )

        try:
            module = getattr(module, _cls)
        except AttributeError:
            raise AttributeError(
                f"Cannot resolve {name} from module {module}. Maybe it moved since you saved this object?"
            )

    try:
        obj = getattr(module, cls)
    except AttributeError:
        global_path = current_context().path
        raise AttributeError(
            f"Cannot resolve {name} from module {module}. Maybe it moved since you saved this object?\n"
            f"This error happened while processing object at path `{global_path}`."
        )
    return obj



from nqxpack._src.registry import stdlib

from nqxpack._src.registry import jax
from nqxpack._src.registry import flax

from nqxpack._src.registry import hydra

from nqxpack._src.registry import netket
from nqxpack._src.registry import netket_operator

from nqxpack._src.registry.versioninfo import VERSION



from functools import lru_cache


import jax

from netket import config as nkconfig
from netket.utils import mpi


@lru_cache
def process_index() -> int:
    """
    Returns the index of this process running NetKet.

    If you are running with experimental sharding, this is
    equivalent to ``jax.process_index()``. If you are running
    with mpi, this is ``nk.utils.mpi.rank``.

    This is an integer between 0 and
    :func:`netket_pro.distributed.process_count()`.
    """

    if nkconfig.netket_experimental_sharding:
        return jax.process_index()
    else:
        return mpi.rank


def is_master_process() -> bool:
    """
    Returns whether the current process is the master process.
    """
    return process_index() == 0



# File containing common commands for NetKet Test infrastructure

from typing import Any

from functools import partial
import os

import pytest

import netket as nk


def _is_true(x):
    if isinstance(x, str):
        xl = x.lower()
        if xl == "1" or x == "true":
            return True
    elif x == 1:
        return True
    else:
        return False


skipif_ci = pytest.mark.skipif(
    _is_true(os.environ.get("CI", False)), reason="Test too slow/broken on CI"
)
"""Use as a decorator to mark a test to be skipped when running on CI.
For example:

Example:
>>> @skipif_ci
>>> def test_my_serial_function():
>>>     your_serial_test()

"""

skipif_mpi = pytest.mark.skipif(nk.utils.mpi.n_nodes > 1, reason="Only run without MPI")
"""Use as a decorator to mark a test to be skipped when running under MPI.
For example:

Example:
>>> @skipif_mpi
>>> def test_my_serial_function():
>>>     your_serial_test()

"""

onlyif_mpi = pytest.mark.skipif(nk.utils.mpi.n_nodes < 2, reason="Only run with MPI")
"""Use as a decorator to mark a test to be executed only when running with at least 2 MPI
nodes.

It can be used in combination with the fixtures defined in test/conftest.py, namely
_mpi_rank, mpi_size and _mpi_comm, that retrieve the number of mpi rank, size and comm
used by netket.

If you need to trick netket into executing some code without MPI when running under MPI,
you can use the class netket_disable_mpi below

Example:

>>> @onlyif_mpi
>>> def test_my_parallel_function():
>>>     x = compute_some_stuff_with_mpi
>>>     with netket_disable_mpi():
>>>         x_serial = compute_some_stuff_withjout_mpi

"""

xfailif_mpi = pytest.mark.xfail(
    nk.utils.mpi.n_nodes > 1, reason="custom_vjp not supports effects."
)
"""Use as a decorator to mark a test to be expected to fail only when running with
at least 2 MPI processes.
"""

skipif_sharding = pytest.mark.skipif(
    nk.config.netket_experimental_sharding, reason="Only run without MPI"
)

skipif_distributed = pytest.mark.skipif(
    nk.utils.mpi.n_nodes > 1 or nk.config.netket_experimental_sharding,
    reason="Skip if distributed",
)


class netket_disable_mpi:
    """
    Temporarily disables MPI functions inside of NetKet, tricking NetKet
    into executing all code on every rank independently.

    Example:

    >>> with netket_disable_mpi():
    >>>     run_code

    """

    def __enter__(self):
        self._orig_nodes = nk.utils.mpi.n_nodes
        nk.utils.mpi.n_nodes = 1
        nk.utils.mpi.mpi.n_nodes = 1
        nk.utils.mpi.primitives.n_nodes = 1

    def __exit__(self, exc_type, exc_value, traceback):
        nk.utils.mpi.n_nodes = self._orig_nodes
        nk.utils.mpi.mpi.n_nodes = self._orig_nodes
        nk.utils.mpi.primitives.n_nodes = self._orig_nodes


class set_config:
    """
    Temporarily changes the value of the configuration `name`.

    Example:

    >>> with set_config("netket_experimental_disable_ode_jit", True):
    >>>     run_code

    """

    def __init__(self, name: str, value: Any):
        self._name = name.upper()
        self._value = value

    def __enter__(self):
        self._orig_value = nk.config.FLAGS[self._name]
        nk.config.update(self._name, self._value)

    def __exit__(self, exc_type, exc_value, traceback):
        nk.config.update(self._name, self._orig_value)


netket_experimental_fft_autocorrelation = partial(
    set_config, "NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION"
)


def hash_for_seed(obj):
    """
    Hash any object into an int that can be used in `np.random.seed`, and does not change between Python sessions.

    Args:
      obj: any object with `repr` defined to show its states.
    """

    bs = repr(obj).encode()
    out = 0
    for b in bs:
        out = (out * 256 + b) % 4294967291  # Output in [0, 2**32 - 1]
    return out


def named_parametrize(argname: str, values: list):
    param_values = [pytest.param(obj, id=f"{argname}={obj}") for obj in values]
    return pytest.mark.parametrize(argname, param_values)



from functools import partial
from types import FunctionType

from nqxpack._src.lib_v1.resolution import (
    _fname,
    _resolve_qualname,
)

CLOSURE_SERIALIZATION_REGISTRY = {}


def is_closure(obj):
    return isinstance(obj, FunctionType) and obj.__closure__ is not None


def _serialize_closure(parent_fun, vars_mapping, closure):
    closure_vars = closure.__closure__

    res = {}
    for i, varname in enumerate(closure.__code__.co_freevars):
        varname = vars_mapping.get(varname, varname)
        res[varname] = closure_vars[i].cell_contents

    return {"_target_": _fname(parent_fun), **res}


def register_closure_simple_serialization(
    parent_fun,
    closure_name,
    vars_mapping={},
    original_qualname=None,
    override: bool = False,
):
    """
    Register the closure resulting from a function to have a simple serialization.

    This is usually enough for jax.nn.initializers, for example. The requirement
    for this to work is that the closure shows has a name of the form
    <parent_fun>.<locals>.<closure_name>.

    To register, you must provide the parent function from which the closure
    was generated, the name of the closure, and optionally a mapping of the closure
    variable names to the names you want to use in the serialization.

    Args:
        parent_fun: The parent function of the closure
        closure_name: The name of the closure
        vars_mapping: A mapping of the closure variable names to the names you want
            to use in the serialization
        original_qualname: The original qualname of the parent function. Jax usually
            hides the ._src part of the qualname, and you must provide it here to
            resolve the original function.
    """
    if original_qualname is not None:
        assert _resolve_qualname(original_qualname) == parent_fun
        qualname = original_qualname
    else:
        qualname = _fname(parent_fun)

    closure_qualname = qualname + ".<locals>." + closure_name

    if closure_qualname in CLOSURE_SERIALIZATION_REGISTRY and not override:
        raise ValueError(
            f"Closure '{closure_qualname}' is already registered for serialization"
        )

    CLOSURE_SERIALIZATION_REGISTRY[closure_qualname] = partial(
        _serialize_closure, parent_fun, vars_mapping
    )
